'''
$Id: GenericUserFolder.py,v 1.25 2000/01/05 02:19:10 zen Exp $
'''

import Globals
from Globals import HTMLFile, MessageDialog
from string import split,join,lower
from Acquisition import Implicit
from Persistence import Persistent
from OFS.SimpleItem import Item
from OFS.Document import Document
from base64 import decodestring, encodestring
from urllib import quote, unquote
from AccessControl.User \
    import BasicUser, BasicUserFolder, _remote_user_mode,nobody
from AccessControl.PermissionRole import _what_not_even_god_should_do
from AccessControl.Role import RoleManager
from OFS.Folder import Folder
from time import time
from sys import exc_info
import types
import sha
import whrandom
import Globals
import threading
from zLOG import LOG,ERROR,WARNING

from User import User # Our user object

# _hashmap contains our authentication tokens
# _hashmap[_p_oid][hash] == (username,lastseentime)
_hashmap = {} 		

def _debug(msg): pass
#def _debug(msg): LOG('GUF',WARNING,msg)

class GenericUserFolder(Folder,BasicUserFolder,Persistent,Implicit,RoleManager):
    '''A Generic User authentication userlist'''

    meta_type='Generic User Folder'
    id       ='acl_users'
    title    ='Generic User Folder'
    icon     ='misc_/GenericUserFolder/Folder_icon'

    isPrincipiaFolderish=1
    isAUserFolder=1
    isAGenericUserFolder=1

    __ac_permissions__=[
	['View management screens',
	    ['cachestats','flushcache','manage_cache',
	     'manage','manage_menu','manage_main'],['Manager',]],
	['Manage users', 
	    ['getUserNames','getUser','getUsers','manage_listUsers',
	     'manage_userInfo'],['Manager',]],
 	['Manage properties',
	    ['manage_edit','manage_properties'],['Manager',]],
	['Can Login and Logout',
	    ['login','logout','loginSuccess' ],['Anonymous','Manager']]]

    manage_options=(
	{'label':'Contents',	'action':'manage_main'},
	{'label':'User List',	'action':'manage_listUsers'},
	{'label':'Cache',	'action':'manage_cache'},
	{'label':'User Info',	'action':'manage_userInfo'},
	{'label':'Properties',	'action':'manage_properties'},
	{'label':'Security',	'action':'manage_access'},
	{'label':'Undo',	'action':'manage_UndoForm'}
    )

    manage_properties=HTMLFile('manage_properties', globals())
    manage_listUsers=HTMLFile('manage_listUsers',globals())
    manage_userInfo=HTMLFile('manage_userInfo',globals())
    manage_cache=HTMLFile('manage_cache',globals())
    
    def __init__(self, cookie_mode=0, REQUEST=None):
        self.cookie_mode=cookie_mode
	self.cookie_expires=900
	self.userlist_expires=900 # Obsolete?
	self.user_expires=900 # Obsolete?

	#self.roles = ['Anonymous',]

	self._init_caches()
    
    def __setstate__(self,state):
	"Initialize on reload from ZODB"
	Folder.__setstate__(self,state)
	self._init_caches()

    def _init_caches(self):
	gufid = self._p_oid
	_debug('gudid is %s' % str(gufid))

	if not _hashmap.has_key(gufid):
	    _hashmap[gufid] = {}

    def cachestats(self):
	'Return meaningless statistics'
	gufid = self._p_oid

	if not _hashmap.has_key(gufid):
	    _hashmap[gufid] = {}

	return (\
	    {'text':	'Items in Authorization Token cache',
	     'value':	len(_hashmap[gufid] or [])},)
		 
    def flushcache(self,REQUEST):
	'Flush the caches'

	gufid = self._p_oid
	if REQUEST and REQUEST.has_key('flushcookies') and REQUEST.flushcookies:
	    _hashmap[gufid] = {}
	    LOG('GUF',0,"Authentication tokens flushed by %s" \
		% REQUEST.AUTHENTICATED_USER.getUserName())
            return MessageDialog(
                   title  ='GenericUserFolder Caches Flushed',
                   message='''GenericUserFolder authentication tokens 
			    have been flushed''',
                   action = './manage_cache')
	
        if REQUEST:
            return MessageDialog(
                   title  ='GenericUserFolder Caches WERE NOT Flushed',
                   message='''GenericUserFolder authentication tokens have 
			<strong>NOT</strong> been flushed, as the safty-valve 
			checkbox was not selected''',
                   action = './manage_cache')
	
    def getUserNames(self,sortem=0):
        '''Returns a list of user names or [] if no users exist'''

	if not hasattr(self,'userList'):
	    LOG('GUF',ERROR,'%s has no userList method' \
		% self.absolute_url())
	    return []

	try:
	    data = self.userList(self,self.REQUEST)

	    if type(data) == types.StringType:
		data = split(data)

	    data = list(data)

	    if len(data) > 0 and type(data[0]) != types.StringType:
		if hasattr(data[0],'username'):
		    for i in range(0,len(data)):
			data[i] = data[i].username
		elif hasattr(data[0],'uname'):
		    for i in range(0,len(data)):
			data[i] = data[i].uname
		
		if type(data[0]) != types.StringType:
		    raise 'Unsupported Type',\
			'userList method returned an unsupported type'

	    if sortem:
		data.sort()
	    return data

	except:
	    LOG('GUF',ERROR, 'An exception was raised by %s (%s)' \
		% (self.userList.absolute_url(),str(exc_info())),'',exc_info())
	    return []


    def getUsers(self):
        """Return a list of user objects or [] if no users exist"""

	data = self.getUserNames()

	userlist   = []

	for name in data:
	    user = User(name,self)
	    user = user.__of__(self)
	    userlist.append(user)

	return userlist


    def getUser(self, name):
        """Return the named user object or None if no such user exists"""

	# Special case for the superuser. Useless unless we can
	# somehow make this the top folder.
	super = self._super	
	if self._isTop() and (name==super.getUserName()):
	    return super

	# Shortcut method for sites with large numbers of users
	#
	if hasattr(self,'userExists'):
	    try:
		if self.userExists(self,self.REQUEST,username=name) == 1:
		    user = User(name,self)
		    return user
		else:
		    return None
	    except:
		LOG('GUF',ERROR,'%s raised an exception (%s)' \
		    % (self.userExists.absolute_url(),str(exc_info())),'',exc_info())

	data = self.getUserNames()
	if name in data:
	    user = User(name,self)
	    return user

	return None


    def _stashhash(self,username,passwd,request):
	'''stash the hash in our _hashmap cache. Set the cookie.'''

	gufid = self._p_oid
	_debug('gufid = %s' % str(gufid))

	response = request['RESPONSE']
	hash = encodestring(sha.new('%s%s%f%f%s' \
	    % (username,passwd,time(),whrandom.random(),str(request))).digest())

	try:
	    hashmap = _hashmap[gufid]
	except KeyError:
	    hashmap = {}
	    _hashmap[gufid] = hashmap

	# The Universe will probably end before a sha hash isn't unique,
	# but call me paranoid
	while hashmap.has_key(hash):
	    hash = encodestring(sha.new('%s%s%f%f%s' \
		% (username,passwd,time(),
		   whrandom.random(),str(request))).digest())

		    # List rather than tuple so its mutable
	hashmap[hash] = [time(),username]

	# TODO: Calculate path so that cookie only gets passed to the
	# acl_users folder's parent folder and below?
	response.setCookie('_gufauth',quote(hash),path='/')

	if request.has_key('URL'):
	    _debug('URL is %s' % request['URL'])
	else:
	    _debug('No URL!')


    def _recurse_validate(self,request,auth,roles,\
		username=None,passwd=None,hash=None):
	'''Try validating locally, or crawl up the tree until
	   we succeed or reach the top'''

	gufid = self._p_oid

	# Check from BasicUserFolder.validate. I don't actually know
	# what it does except that in some circumstances it will deny
	# access (better safe than sorry)
	if roles is _what_not_even_god_should_do:
	    _debug('_what_not_even_god_should_do request')
	    request.response.notFoundError()

	parents = []
	if request.has_key('PARENTS'):
	    parents=request['PARENTS']
	    parent=parents[0]
	else:
	    parent=self.aq_parent

	# If we have a password, then someone has just submitted the login
	# form
	if passwd is not None:
	    user = self.getUser(username)
	    if user is not None and user.authenticate(passwd,request) \
		    and user.allowed(parent,roles):
		self._stashhash(username,passwd,request)
		user = user.__of__(self)
		return user

	# If we have a hash, then we are confirming the validity of a cookie
	if hash is not None:

	    try:
		hashmap = _hashmap[gufid]
	    except KeyError:
		hashmap = {}
		_hashmap[gufid] = hashmap
		    
	    # If we are using cookie authentication, check our hashmap
	    # stash. We do it this way so that our hashes are shared
	    # between threads (otherwise, you would have to authenticate
	    # once for each thread that handles one of your requests).
	    if hashmap.has_key(hash):

		hashitem = hashmap[hash]

		if time() > hashitem[0] + self.cookie_expires:
		    # Our cookie has expired!
		    del hashmap[hash]
		else:
		    # Update our lastseen time
		    hashitem[0] = time()
		    username = hashmap[hash][1]

		    user = self.getUser(username)
		    if user is not None and user.allowed(parent,roles):
			user = user.__of__(self)
			return user

	# if we have been authenticated using basic authentication
	# using a higher acl_users folder, we can try authenticating
	# in the guf with the same username & password. 
	if auth and lower(auth[:6]) == 'basic ':
	    username,passwd = tuple(split(
		decodestring(split(auth)[-1]), ':', 1))
	    user = self.getUser(username)
	    if user is not None and user.authenticate(passwd,request) \
		    and user.allowed(parent,roles):
		self._stashhash(username,passwd,request)
		user = user.__of__(self)
		return user

	# We have failed to validate here. We now need to climb the
	# tree and try validating with other acl_users folders.
	# Although Zope also does this, we can't use it: The only way
	# of displaying a nice HTML login screen is by throwing an exception
	# out of the validate method, which has the side effect of causing
	# 'higher' acl_users folders to not be checked.
	#
	# Other UserFolder products that handle cookies (etcUserFolder and
	# UserDB) will just return 'None' from this function if auth is
	# not none - however, this doesn't work if the user is already
	# authenticated with a higher level acl_users folder

	_debug('Failed to authenticate with %s - crawling up' \
	    % self.absolute_url())

	lastuserfolder = self
	if hasattr(parent,'aq_parent'):
	    c = parent.aq_parent
	    while c is not None:

		if hasattr(c,'aq_explicit') \
		    and hasattr(c.aq_explicit,'acl_users'):
		    
		    u = c.aq_explicit.acl_users

		    # This check is probably just paranoia
		    if u.absolute_url() != lastuserfolder.absolute_url():
			lastuserfolder = u

			if hasattr(u,'isAGenericUserFolder'):
			    # If we find a GUF above, we won't need to
			    # crawl up any further as its validate method
			    # will do it for us
			    user = u.validate(request,auth,roles)
			    if user:
				return user
			    else:
				return None

			if hasattr(u,'isAUserFolder') and u.isAUserFolder:
			    user = u.validate(request,auth,roles)

			    if user and user.allowed(parent,roles):
				_debug('Authenticated %s with %s' \
				    % (user.getUserName(),u.absolute_url()))
				return user
			    else:
				_debug('Failed Authentication with %s' \
				    % u.absolute_url())
				
		if hasattr(c,'aq_parent'):
		    c = c.aq_parent
		else:
		    c = None

	# Patch From: Chris Allen <chris_allen@aurema.com>
        # Allow anonymous access to public objects.
	nobody = self._nobody
	if nobody.allowed(parent, roles):
	    obj = nobody.__of__(self)
	    return obj
	
	return None


    def _cookie_validate(self,request,auth,roles=None):
        response=request['RESPONSE']

	cname = '_gufauth' 
	nameparm = cname + '_name'
	passparm = cname + '_password'

	username = None
	password = None
	hash = None

	gufid = self._p_oid

	if request.form.has_key(nameparm):
	    #
	    # Pull username & cleartext from form input.
	    # Ensure cleartext isn't left in the REQUEST for
	    # malicious DTML
	    #
	    # Hmm... I can delete from request.form, but not request...
	    #
	    username = request[nameparm]
	    request.form[nameparm] = 'This space intentionally blank'
	    request[nameparm] = 'This space intentionally blank'

	if request.form.has_key(passparm):
	    password = request[passparm]
	    request.form[passparm] = 'This space intentionally blank'
	    request[passparm] = 'This space intentionally blank'
	
	e = request.environ
	if _remote_user_mode and e.has_key('REMOTE_USER'):
	    username = e['REMOTE_USER']
    
	if request.cookies.has_key(cname):
	    # Pull username & hash from cookie	
	    hash = unquote(request.cookies[cname])  

	if username is not None or password is not None:

	    # If user is relogging in again, nuke their
	    # old cookie right now
	    if hash is not None:
		try: del _hashmap[gufid][hash]
		except KeyError: pass

	    #
	    # Try validating with username & password, or fall through to
	    # using 'auth' on our parent folders.
	    #
	    user = self._recurse_validate(request,auth,roles,username,password)

	    if user is not None:
		return user

	elif hash is not None:
	    user = self._recurse_validate(request,auth,roles,hash=hash)

	    if user: 
		return user
	    else:
		response.expireCookie('_gufauth', path='/')

	else: # We are authenticated using a higher acl_users folder?
	    user = self._recurse_validate(request,auth,roles)
	    if user:
		return user

	# We can log a message here, as we already have tried to authenticate
	# against 'higher' acl_users folders. A nice side effect of having
	# to do this work ourselves. This log message might mean a bad
	# password, or a valid user trying to access something they have
	# no rights to see.
	if username is not None and username != '':
	    LOG('GUF',0,'Failed authorization for user %s (%s)' \
		% (username,self.absolute_url()))

	if hasattr(self,'docLoginSuccess'):
	    raise 'LoginRequired', self.docLogin(self, request,
		loginSuccessURL=self.absolute_url() + '/loginSuccess')
	else:
	    raise 'LoginRequired', self.docLogin(self, request,
		loginSuccessURL='')

    def _basic_validate(self,request,auth,roles):
	# Similar to BasicUserFolder.validate, except don't try
	# "If no authorization, only a user with a domain spec and no passwd 
	#  or nobody can match" section, as this is real bad with large user
	# bases

        if roles is _what_not_even_god_should_do:
            request.response.notFoundError()

        parents=request.get('PARENTS', [])
        if not parents:
            parent=self.aq_parent
        else: parent=parents[0]

        # Only do basic authentication
        if lower(auth[:6])!='basic ':
            return None
        name,password=tuple(split(decodestring(split(auth)[-1]), ':', 1))

        # Check for superuser
        super=self._super
        if self._isTop() and (name==super.getUserName()) and \
        super.authenticate(password, request):
            return super

        # Try to get user
        user=self.getUser(name)
        if user is None:
            return None

        # We need the user to be able to acquire!
        user=user.__of__(self)

        # Try to authorize user
        if not user.allowed(parent, roles):
            return None

        # Try to authenticate user
        if user.authenticate(password, request):
            return user

        return None


    def validate(self,request,auth='',roles=None):
	# No docstring for a reason

	# Make sure auth is a string
	if auth is None:
	    auth = ''

	if request.has_key('URL'):
	    _debug('URL is %s in validate' % request['URL'])
	else:
	    _debug('No URL in validate')

        if self.cookie_mode:
            u = self._cookie_validate(request, auth, roles)
        else:
	    u = self._basic_validate(request,auth,roles)

	if u is None:
	    _debug("Returning None from %s validate (roles=%s)" \
		% (self.absolute_url(),str(roles)))
	else:
	    _debug("Returning %s from %s validate (roles=%s)" \
		% (u.getUserName(),self.absolute_url(),str(roles)))
	return u
	
    def __len__(self):
        return len(self.getUserNames())

    def manage_edit(self, cookie_mode, userlist_expires,user_expires,
	    cookie_expires, REQUEST=None):
	"""Change properties"""

        self.cookie_mode=cookie_mode
	self.userlist_expires=userlist_expires 	# Obsolete?
	self.user_expires=user_expires		# Obsolete?
	self.cookie_expires=cookie_expires

        if REQUEST:
            return MessageDialog(
                   title  ='GenericUserFolder Changed',
                   message='GenericUserFolder properties have been updated',
                   action =REQUEST['URL1']+'/manage_properties',
                   target ='manage_main')

    # Multiple inheritance wierdness
    #
    def manage_beforeDelete(self,item,container):
	Folder.manage_beforeDelete(self,item,container)
	BasicUserFolder.manage_beforeDelete(self,item,container)

    def manage_afterAdd(self,item,container):
	Folder.manage_afterAdd(self,item,container)
	BasicUserFolder.manage_afterAdd(self,item,container)

    def _killcookie(self,request):
	'Helper method to login and logoff'
	if not self.cookie_mode: raise 'Method only works in cookie mode'
	cname='_gufauth'
	if request.cookies.has_key(cname):
	    # Pull username & hash from cookie	
	    try:
		hash = unquote(request.cookies[cname])  
		del _hashmap[self._p_oid][hash]
	    except KeyError: pass
	request['RESPONSE'].expireCookie(cname,path='/')

    def login(self, REQUEST):
	"""Login"""
	self._killcookie(REQUEST)
        return self.docLogin(self, REQUEST,
	    loginSuccessURL=self.absolute_url() + '/loginSuccess')

    def logout(self, REQUEST):
        """Logout"""
	self._killcookie(REQUEST)
        return self.docLogout(self, REQUEST)

    def loginSuccess(self,REQUEST):
	"""Successful Login"""
	return self.docLoginSuccess(self,REQUEST)

    def _templateFile(self,file,title='',template='dumb'):
	t = HTMLFile('templates/%s/%s' % (template,file),globals())
	self.manage_addDTMLMethod(file,title,t)
    
    def _templateReset(self, template='dumb'):
	self._templateFile('userList','List Users',template)
	self._templateFile('userRoles',"Return a user's roles",template)
	self._templateFile('userDomains','Return domains a user can login from',
	    template)
	self._templateFile('userAuthenticate','Authenticate a user',template)
	self._templateFile('docLogin','Cookie mode login screen',template)
	self._templateFile('docLogout','Cookie mode logout screen',template)
	self._templateFile('docLoginSuccess','Successful Authentication Screen',
	    template)


# Magic bullet?
Globals.default__class_init__(GenericUserFolder)

manage_addGenericUserFolderForm=HTMLFile('addGenericUserFolder', globals())

def manage_addGenericUserFolder(self,cookie_mode=0,
				template='dumb',REQUEST=None):
    'Create a GenericUserFolder object'

    if hasattr(self.aq_base, 'acl_users'):
        return MessageDialog(
               title  ='Item Exists',
               message='This object already contains a User Folder',
               action ='%s/manage_main' % REQUEST['PARENT_URL'])

    ob=GenericUserFolder(cookie_mode, REQUEST)
    self._setObject('acl_users', ob)
    self.__allow_groups__=self.acl_users

    ob._templateReset(template)

    if REQUEST: return self.manage_main(self,REQUEST,update_menu=1)

